// Copyright 2022 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package badges

import (
	"github.com/rs/zerolog/log"

	"go.woodpecker-ci.org/woodpecker/v3/server/model"
)

var (
	// Status labels.
	badgeStatusSuccess = "success"
	badgeStatusFailure = "failure"
	badgeStatusStarted = "started"
	badgeStatusError   = "error"
	badgeStatusNone    = "none"
)

func getBadgeStatusLabelAndColor(status *model.StatusValue) (string, Color) {
	if status == nil {
		return badgeStatusNone, ColorGray
	}

	switch *status {
	case model.StatusSuccess:
		return badgeStatusSuccess, ColorGreen
	case model.StatusFailure:
		return badgeStatusFailure, ColorRed
	case model.StatusPending, model.StatusRunning:
		return badgeStatusStarted, ColorYellow
	case model.StatusError, model.StatusKilled:
		return badgeStatusError, ColorGray
	default:
		return badgeStatusNone, ColorGray
	}
}

// Generate an SVG badge based on a pipeline.
func Generate(name string, status *model.StatusValue) (string, error) {
	label, color := getBadgeStatusLabelAndColor(status)
	bytes, err := RenderBytes(name, label, color)
	if err != nil {
		log.Warn().Err(err).Msg("could not render badge")
		return "", err
	}
	return string(bytes), nil
}
