// Copyright 2023 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package linter

import (
	pipeline_errors "go.woodpecker-ci.org/woodpecker/v3/pipeline/errors"
)

func newLinterError(message, file, field string, isWarning bool) *pipeline_errors.PipelineError {
	return &pipeline_errors.PipelineError{
		Type:      pipeline_errors.PipelineErrorTypeLinter,
		Message:   message,
		Data:      &pipeline_errors.LinterErrorData{File: file, Field: field},
		IsWarning: isWarning,
	}
}
