// Copyright 2021 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package main

import (
	"context"
	"os"

	"github.com/rs/zerolog/log"

	"go.woodpecker-ci.org/woodpecker/v3/shared/dot_env"
	"go.woodpecker-ci.org/woodpecker/v3/shared/utils"
)

func main() {
	dot_env.Load()

	ctx := utils.WithContextSigtermCallback(context.Background(), func() {
		log.Info().Msg("termination signal is received, terminate cli")
	})

	app := newApp()
	if err := app.Run(ctx, os.Args); err != nil {
		log.Fatal().Err(err).Msg("error running cli") //nolint:forbidigo
	}
}
