// Copyright 2024 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package runtime

import (
	"context"
	"sync"
	"time"
)

const shutdownTimeout = time.Second * 5

var (
	shutdownCtx     context.Context
	shutdownCtxLock sync.Mutex
)

// GetShutdownCtx returns a context that is valid for shutdownTimeout after the
// first call. It is used as a fallback cleanup context when the runner context
// is already canceled.
func GetShutdownCtx() context.Context {
	shutdownCtxLock.Lock()
	defer shutdownCtxLock.Unlock()
	if shutdownCtx == nil {
		shutdownCtx, _ = context.WithTimeout(context.Background(), shutdownTimeout) //nolint:govet
	}
	return shutdownCtx
}
