// Copyright 2023 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package migration

import (
	"src.techknowlogick.com/xormigrate"
	"xorm.io/xorm"
)

var legacyToXormigrate = xormigrate.Migration{
	ID: "legacy-to-xormigrate",
	MigrateSession: func(sess *xorm.Session) error {
		type migrations struct {
			Name string `xorm:"UNIQUE"`
		}

		var mig []*migrations
		if err := sess.Find(&mig); err != nil {
			return err
		}
		for _, m := range mig {
			if _, err := sess.Insert(&xormigrate.Migration{
				ID: m.Name,
			}); err != nil {
				return err
			}
		}

		return sess.DropTable("migrations")
	},
}
