// Copyright 2023 The narqo/go-badge Authors. All rights reserved.
// SPDX-License-Identifier: MIT.

package badges

// Color represents color of the badge.
type Color string

// Standard colors.
const (
	ColorGreen  = "#44cc11"
	ColorYellow = "#dfb317"
	ColorRed    = "#e05d44"
	ColorGray   = "#9f9f9f"
)
