// Copyright 2026 Google LLC
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

// Package random provides random number generation functions.
package random

import "crypto/rand"

// MustRand fills b with random bytes.
//
// Panics if the random number generator fails.
func MustRand(b []byte) {
	if _, err := rand.Read(b); err != nil {
		panic(err) // out of randomness, should never happen
	}
}
