// Copyright 2026 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//      http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package dot_env

import (
	"fmt"
	"os"

	"github.com/joho/godotenv"
)

const dotEnv = ".env"

func Load() {
	if _, err := os.Stat(dotEnv); os.IsNotExist(err) {
		return
	}

	if err := godotenv.Load(dotEnv); err != nil {
		fmt.Fprintf(os.Stderr, "Error could not load %q: %s", dotEnv, err)
		os.Exit(1)
	}
}
