// Copyright 2026 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package migration

import (
	"src.techknowlogick.com/xormigrate"
	"xorm.io/xorm"

	"go.woodpecker-ci.org/woodpecker/v3/server/model"
)

var addCronField = xormigrate.Migration{
	ID: "add-cron-field",
	MigrateSession: func(sess *xorm.Session) error {
		type pipelines struct {
			ID int64 `xorm:"pk autoincr 'id'"`

			// new cron field
			Cron string `xorm:"cron"`
		}

		if err := sess.Sync(new(pipelines)); err != nil {
			return err
		}

		_, err := sess.Exec("UPDATE pipelines SET cron = sender, sender = '', message = '' WHERE event = ?", model.EventCron)
		return err
	},
}
