// Code generated by go-swagger; DO NOT EDIT.

package models

// This file was generated by the swagger tool.
// Editing this file might prove futile when you re-run the swagger generate command

import (
	"context"

	"github.com/go-openapi/strfmt"
	"github.com/go-openapi/swag"
)

// QuotaRuleInfo QuotaRuleInfo contains information about a quota rule
//
// swagger:model QuotaRuleInfo
type QuotaRuleInfo struct {

	// The limit set by the rule
	Limit int64 `json:"limit,omitempty"`

	// Name of the rule (only shown to admins)
	Name string `json:"name,omitempty"`

	// Subjects the rule affects
	Subjects []string `json:"subjects"`
}

// Validate validates this quota rule info
func (m *QuotaRuleInfo) Validate(formats strfmt.Registry) error {
	return nil
}

// ContextValidate validates this quota rule info based on context it is used
func (m *QuotaRuleInfo) ContextValidate(ctx context.Context, formats strfmt.Registry) error {
	return nil
}

// MarshalBinary interface implementation
func (m *QuotaRuleInfo) MarshalBinary() ([]byte, error) {
	if m == nil {
		return nil, nil
	}
	return swag.WriteJSON(m)
}

// UnmarshalBinary interface implementation
func (m *QuotaRuleInfo) UnmarshalBinary(b []byte) error {
	var res QuotaRuleInfo
	if err := swag.ReadJSON(b, &res); err != nil {
		return err
	}
	*m = res
	return nil
}
