// Copyright 2024 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package output

import (
	"errors"
	"strings"
)

var ErrOutputOptionRequired = errors.New("output option required")

func ParseOutputOptions(out string) (string, []string) {
	out, opt, found := strings.Cut(out, "=")

	if !found {
		return out, nil
	}

	var optList []string

	if opt != "" {
		optList = strings.Split(opt, ",")
	}

	return out, optList
}
