// Copyright 2023 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package yaml

import (
	"codeberg.org/6543/xyaml/v2"

	"go.woodpecker-ci.org/woodpecker/v3/pipeline/frontend/yaml/types"
)

const (
	// Set to allow deeply nested k8s backend_options for node affinity.
	maxMergeDepth uint16 = 15
)

// ParseBytes parses the configuration from bytes b.
func ParseBytes(b []byte) (*types.Workflow, error) {
	parser := xyaml.NewParser(xyaml.WithDepth(maxMergeDepth))
	out := new(types.Workflow)
	err := parser.Unmarshal(b, out)
	if err != nil {
		return nil, err
	}
	return out, nil
}

// ParseString parses the configuration from string s.
func ParseString(s string) (*types.Workflow, error) {
	return ParseBytes(
		[]byte(s),
	)
}
