// Copyright 2026 Woodpecker Authors
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.

package rpc

import "errors"

var (
	ErrAgentIllegalWorkflowReRunStateChange = errors.New("workflow was already marked as finished")
	ErrAgentIllegalWorkflowRun              = errors.New("workflow is currently in blocked state")

	ErrAgentIllegalLogStreaming = errors.New("agent can not append logs to a step that is marked not running")

	ErrAgentIllegalRepo            = errors.New("agent is not allowed to interact with repo")
	ErrAgentIllegalWorkflowAgentID = errors.New("agent is not allowed to interact with workflow where it's id is not already locked")
)
